@@ -12,6 +12,7 @@ import sys
 import uuid
 import warnings
 from difflib import get_close_matches
+from functools import partial
 from io import StringIO
 from logging import getLogger
 from threading import Lock
@@ -23,8 +24,10 @@ from typing import (
     Generator,
     Iterable,
     List,
+    NamedTuple,
     Optional,
     Tuple,
+    Type,
     Union,
 )
 
@@ -38,13 +41,10 @@ from .auth_okta import AuthByOkta
 from .auth_usrpwdmfa import AuthByUsrPwdMfa
 from .auth_webbrowser import AuthByWebBrowser
 from .bind_upload_agent import BindUploadError
-from .chunk_downloader import (
-    DEFAULT_CLIENT_PREFETCH_THREADS,
-    MAX_CLIENT_PREFETCH_THREADS,
-    SnowflakeChunkDownloader,
-)
 from .compat import IS_LINUX, IS_WINDOWS, quote, urlencode
 from .constants import (
+    DEFAULT_S3_CONNECTION_POOL_SIZE,
+    MAX_S3_CONNECTION_POOL_SIZE,
     PARAMETER_AUTOCOMMIT,
     PARAMETER_CLIENT_PREFETCH_THREADS,
     PARAMETER_CLIENT_REQUEST_MFA_TOKEN,
@@ -95,13 +95,12 @@ from .network import (
 from .sqlstate import SQLSTATE_CONNECTION_NOT_EXISTS, SQLSTATE_FEATURE_NOT_SUPPORTED
 from .telemetry import TelemetryClient
 from .telemetry_oob import TelemetryService
-from .time_util import (
-    DEFAULT_MASTER_VALIDITY_IN_SECONDS,
-    HeartBeatTimer,
-    get_time_millis,
-)
+from .time_util import HeartBeatTimer, get_time_millis
 from .util_text import construct_hostname, parse_account, split_statements
 
+DEFAULT_CLIENT_PREFETCH_THREADS = 4
+MAX_CLIENT_PREFETCH_THREADS = 10
+
 
 def DefaultConverterClass():
     if IS_WINDOWS:
@@ -159,16 +158,16 @@ DEFAULT_CONFIGURATION = {
     "inject_client_pause": (0, int),  # snowflake internal
     "session_parameters": (None, (type(None), dict)),  # snowflake session parameters
     "autocommit": (None, (type(None), bool)),  # snowflake
-    "client_session_keep_alive": (False, bool),  # snowflake
+    "client_session_keep_alive": (None, (type(None), bool)),  # snowflake
     "client_session_keep_alive_heartbeat_frequency": (
         None,
         (type(None), int),
     ),  # snowflake
     "client_prefetch_threads": (4, int),  # snowflake
+    "s3_connection_pool_s": (DEFAULT_S3_CONNECTION_POOL_SIZE, int),  # boto3 pool size
     "numpy": (False, bool),  # snowflake
     "ocsp_response_cache_filename": (None, (type(None), str)),  # snowflake internal
     "converter_class": (DefaultConverterClass(), SnowflakeConverter),
-    "chunk_downloader_class": (SnowflakeChunkDownloader, object),  # snowflake internal
     "validate_default_parameters": (False, bool),  # snowflake
     "probe_connection": (False, bool),  # snowflake
     "paramstyle": (None, (type(None), str)),  # standard/snowflake
@@ -191,6 +190,9 @@ DEFAULT_CONFIGURATION = {
         False,
         bool,
     ),  # only use regional url when the param is set
+    # Allows cursors to be re-iterable
+    "reuse_results": (False, bool),
+    "use_new_put_get": (True, bool),
 }
 
 APPLICATION_RE = re.compile(r"[\w\d_]+")
@@ -205,6 +207,13 @@ strptime("20150102030405", "%Y%m%d%H%M%S")
 logger = getLogger(__name__)
 
 
+class TypeAndBinding(NamedTuple):
+    """Stores the type name and the Snowflake binding."""
+
+    type: str
+    binding: Optional[str]
+
+
 class SnowflakeConnection(object):
     """Implementation of the connection object for the Snowflake Database.
 
@@ -233,6 +242,7 @@ class SnowflakeConnection(object):
         network_timeout: Network timeout. Used for general purpose.
         client_session_keepalive: Whether to keep connection alive by issuing a heartbeat.
         client_session_keep_alive_heartbeat_frequency: Heartbeat frequency to keep connection alive in seconds.
+        s3_connection_pool_size: Size of connection pool for S3 boto driver.  Default is 10.
         client_prefetch_threads: Number of threads to download the result set.
         rest: Snowflake REST API object. Internal use only. Maybe removed in a later release.
         application: Application name to communicate with Snowflake as. By default, this is "PythonConnector".
@@ -371,21 +381,26 @@ class SnowflakeConnection(object):
 
     @client_session_keep_alive.setter
     def client_session_keep_alive(self, value):
-        self._client_session_keep_alive = True if value else False
+        self._client_session_keep_alive = value
 
     @property
     def client_session_keep_alive_heartbeat_frequency(self):
-        return (
-            self._client_session_keep_alive_heartbeat_frequency
-            if self._client_session_keep_alive_heartbeat_frequency
-            else DEFAULT_MASTER_VALIDITY_IN_SECONDS / 16
-        )
+        return self._client_session_keep_alive_heartbeat_frequency
 
     @client_session_keep_alive_heartbeat_frequency.setter
     def client_session_keep_alive_heartbeat_frequency(self, value):
         self._client_session_keep_alive_heartbeat_frequency = value
         self._validate_client_session_keep_alive_heartbeat_frequency()
 
+    @property
+    def _s3_connection_pool_size(self) -> int:
+        return self._s3_connection_pool_s
+
+    @_s3_connection_pool_size.setter
+    def _s3_connection_pool_size(self, value: int) -> None:
+        self._s3_connection_pool_s = value
+        self._validate_s3_connection_pool_size()
+
     @property
     def client_prefetch_threads(self):
         return (
@@ -566,7 +581,9 @@ class SnowflakeConnection(object):
         """Rolls back the current transaction."""
         self.cursor().execute("ROLLBACK")
 
-    def cursor(self, cursor_class: SnowflakeCursor = SnowflakeCursor):
+    def cursor(
+        self, cursor_class: Type[SnowflakeCursor] = SnowflakeCursor
+    ) -> SnowflakeCursor:
         """Creates a cursor object. Each statement will be executed in a new cursor object."""
         logger.debug("cursor")
         if not self.rest:
@@ -697,10 +714,12 @@ class SnowflakeConnection(object):
                 PARAMETER_CLIENT_VALIDATE_DEFAULT_PARAMETERS
             ] = True
 
-        if self.client_session_keep_alive:
-            self._session_parameters[PARAMETER_CLIENT_SESSION_KEEP_ALIVE] = True
+        if self.client_session_keep_alive is not None:
+            self._session_parameters[
+                PARAMETER_CLIENT_SESSION_KEEP_ALIVE
+            ] = self._client_session_keep_alive
 
-        if self.client_session_keep_alive_heartbeat_frequency:
+        if self.client_session_keep_alive_heartbeat_frequency is not None:
             self._session_parameters[
                 PARAMETER_CLIENT_SESSION_KEEP_ALIVE_HEARTBEAT_FREQUENCY
             ] = self._validate_client_session_keep_alive_heartbeat_frequency()
@@ -728,6 +747,9 @@ class SnowflakeConnection(object):
         self._password = None  # ensure password won't persist
 
         if self.client_session_keep_alive:
+            # This will be called after the heartbeat frequency has actually been set.
+            # By this point it should have been decided if the heartbeat has to be enabled
+            # and what would the heartbeat frequency be
             self._add_heartbeat()
 
     def __preprocess_auth_instance(self, auth_instance):
@@ -905,6 +927,7 @@ class SnowflakeConnection(object):
         is_file_transfer: bool = False,
         statement_params: Optional[Dict[str, str]] = None,
         is_internal: bool = False,
+        describe_only: bool = False,
         _no_results: bool = False,
         _update_current_object: bool = True,
     ):
@@ -920,6 +943,8 @@ class SnowflakeConnection(object):
             data["parameters"] = statement_params
         if is_internal:
             data["isInternal"] = is_internal
+        if describe_only:
+            data["describeOnly"] = describe_only
         if binding_stage is not None:
             # binding stage for bulk array binding
             data["bindStage"] = binding_stage
@@ -1034,6 +1059,45 @@ class SnowflakeConnection(object):
             raise BindUploadError from exc
         return res
 
+    def _get_snowflake_type_and_binding(
+        self,
+        cursor: Optional["SnowflakeCursor"],
+        v: Union[Tuple[str, Any], Any],
+    ) -> TypeAndBinding:
+        if isinstance(v, tuple):
+            if len(v) != 2:
+                Error.errorhandler_wrapper(
+                    self,
+                    cursor,
+                    ProgrammingError,
+                    {
+                        "msg": "Binding parameters must be a list "
+                        "where one element is a single value or "
+                        "a pair of Snowflake datatype and a value",
+                        "errno": ER_FAILED_PROCESSING_QMARK,
+                    },
+                )
+            snowflake_type, v = v
+        else:
+            snowflake_type = self.converter.snowflake_type(v)
+            if snowflake_type is None:
+                Error.errorhandler_wrapper(
+                    self,
+                    cursor,
+                    ProgrammingError,
+                    {
+                        "msg": "Python data type [{}] cannot be "
+                        "automatically mapped to Snowflake data "
+                        "type. Specify the snowflake data type "
+                        "explicitly.".format(v.__class__.__name__.lower()),
+                        "errno": ER_NOT_IMPLICITY_SNOWFLAKE_DATATYPE,
+                    },
+                )
+        return TypeAndBinding(
+            snowflake_type,
+            self.converter.to_snowflake_bindings(snowflake_type, v),
+        )
+
     # TODO we could probably rework this to not make dicts like this: {'1': 'value', '2': '13'}
     def _process_params_qmarks(
         self,
@@ -1043,6 +1107,7 @@ class SnowflakeConnection(object):
         if not params:
             return None
         processed_params = {}
+
         if not isinstance(params, (list, tuple)):
             errorvalue = {
                 "msg": "Binding parameters must be a list: {}".format(params),
@@ -1050,49 +1115,26 @@ class SnowflakeConnection(object):
             }
             Error.errorhandler_wrapper(self, cursor, ProgrammingError, errorvalue)
 
+        get_type_and_binding = partial(self._get_snowflake_type_and_binding, cursor)
+
         for idx, v in enumerate(params):
-            if isinstance(v, tuple):
-                if len(v) != 2:
-                    Error.errorhandler_wrapper(
-                        self,
-                        cursor,
-                        ProgrammingError,
-                        {
-                            "msg": "Binding parameters must be a list "
-                            "where one element is a single value or "
-                            "a pair of Snowflake datatype and a value",
-                            "errno": ER_FAILED_PROCESSING_QMARK,
-                        },
-                    )
+            if isinstance(v, list):
+                snowflake_type = self.converter.snowflake_type(v)
+                all_param_data = list(map(get_type_and_binding, v))
+                first_type = all_param_data[0].type
+                # if all elements have the same snowflake type, update snowflake_type
+                if all(param_data.type == first_type for param_data in all_param_data):
+                    snowflake_type = first_type
                 processed_params[str(idx + 1)] = {
-                    "type": v[0],
-                    "value": self.converter.to_snowflake_bindings(v[0], v[1]),
+                    "type": snowflake_type,
+                    "value": [param_data.binding for param_data in all_param_data],
                 }
             else:
-                snowflake_type = self.converter.snowflake_type(v)
-                if snowflake_type is None:
-                    Error.errorhandler_wrapper(
-                        self,
-                        cursor,
-                        ProgrammingError,
-                        {
-                            "msg": "Python data type [{}] cannot be "
-                            "automatically mapped to Snowflake data "
-                            "type. Specify the snowflake data type "
-                            "explicitly.".format(v.__class__.__name__.lower()),
-                            "errno": ER_NOT_IMPLICITY_SNOWFLAKE_DATATYPE,
-                        },
-                    )
-                if isinstance(v, list):
-                    vv = [
-                        self.converter.to_snowflake_bindings(
-                            self.converter.snowflake_type(v0), v0
-                        )
-                        for v0 in v
-                    ]
-                else:
-                    vv = self.converter.to_snowflake_bindings(snowflake_type, v)
-                processed_params[str(idx + 1)] = {"type": snowflake_type, "value": vv}
+                snowflake_type, snowflake_binding = get_type_and_binding(v)
+                processed_params[str(idx + 1)] = {
+                    "type": snowflake_type,
+                    "value": snowflake_binding,
+                }
         if logger.getEffectiveLevel() <= logging.DEBUG:
             for k, v in processed_params.items():
                 logger.debug("idx: %s, type: %s", k, v.get("type"))
@@ -1203,7 +1245,10 @@ class SnowflakeConnection(object):
         """Validate and return heartbeat frequency in seconds."""
         real_max = int(self.rest.master_validity_in_seconds / 4)
         real_min = int(real_max / 4)
-        if self.client_session_keep_alive_heartbeat_frequency > real_max:
+        if self.client_session_keep_alive_heartbeat_frequency is None:
+            # This is an unlikely scenario but covering it just in case.
+            self._client_session_keep_alive_heartbeat_frequency = real_min
+        elif self.client_session_keep_alive_heartbeat_frequency > real_max:
             self._client_session_keep_alive_heartbeat_frequency = real_max
         elif self.client_session_keep_alive_heartbeat_frequency < real_min:
             self._client_session_keep_alive_heartbeat_frequency = real_min
@@ -1214,6 +1259,13 @@ class SnowflakeConnection(object):
         )
         return self.client_session_keep_alive_heartbeat_frequency
 
+    def _validate_s3_connection_pool_size(self) -> None:
+        if self._s3_connection_pool_s <= 0:
+            self._s3_connection_pool_s = 1
+        elif self._s3_connection_pool_s > MAX_S3_CONNECTION_POOL_SIZE:
+            self._s3_connection_pool_s = MAX_S3_CONNECTION_POOL_SIZE
+        self._s3_connection_pool_s = int(self._s3_connection_pool_s)
+
     def _validate_client_prefetch_threads(self):
         if self.client_prefetch_threads <= 0:
             self._client_prefetch_threads = 1
@@ -1239,8 +1291,15 @@ class SnowflakeConnection(object):
                 else:
                     TelemetryService.get_instance().disable()
             elif PARAMETER_CLIENT_SESSION_KEEP_ALIVE == name:
-                self.client_session_keep_alive = value
-            elif PARAMETER_CLIENT_SESSION_KEEP_ALIVE_HEARTBEAT_FREQUENCY == name:
+                # Only set if the local config is None.
+                # Always give preference to user config.
+                if self.client_session_keep_alive is None:
+                    self.client_session_keep_alive = value
+            elif (
+                PARAMETER_CLIENT_SESSION_KEEP_ALIVE_HEARTBEAT_FREQUENCY == name
+                and self.client_session_keep_alive_heartbeat_frequency is None
+            ):
+                # Only set if local value hasn't been set already.
                 self.client_session_keep_alive_heartbeat_frequency = value
             elif PARAMETER_SERVICE_NAME == name:
                 self.service_name = value
